#include "ffinterpreter.hh"
#include <iomanip> 
#include <cctype>
#include <stdexcept>
#include <string>

// f*ckf*ck interpreter by Fredrik Arnerup

namespace {
  class EofFailure {};
  int get(std::istream &in) {
    if(!in)
      throw EofFailure();
    return in.get();
  }
}

FFInterpreter::FFInterpreter(std::istream &input, std::ostream &output, 
			     std::ostream &error) :
  in(input), out(output), err(error), ip(0), inside_comment(false),
  last_instruction(FUCK)
{}
 
void FFInterpreter::run() {
  try {
    Program::const_iterator ip = program.begin();
    while(ip != program.end()) {
      execute_instruction(ip);
    }
  } catch (EofFailure) {
    // eof
  }
}

void FFInterpreter::parse(std::istream &src) {
  try {
    for(;;) {
      // print_instruction(out, get_instruction(src), false);
      program.push_back(get_instruction(src));
    }
  } catch (EofFailure) {
    // eof
  }
}

FFInterpreter::instruction FFInterpreter::get_instruction(std::istream &src) {
  std::string firstlast;
  int count = 0;
  for(;;) {
    int c = tolower(get(src)); // case insensitive

    if(inside_comment) { // ignore comment
      if(c == '*' && src.peek() == '/') {
	get(src);
	inside_comment = false;
      }
      continue;
    }
      
    if(c == '/' && src.peek() == '*') { // begin comment
      get(src);
      inside_comment = true;
      continue;
    }

    if(c == '!') // repeat last
      return last_instruction;

    if(c == '*') {
      if(count != 1 && count != 2)
	throw std::runtime_error("'*' is only allowed as the second or "
				 "third f*cking character "
				 "of an instruction");
      count++;
    }

    if(isalpha(c)) {
      if(count == 0 || count == 3)
	firstlast += c;
      count++;

      if(count == 4) {
	if(firstlast == "fk") last_instruction = FUCK;
	else if(firstlast == "sg") last_instruction = SHAG;
	else if(firstlast == "bb") last_instruction = BOOB;
	else if(firstlast == "ts") last_instruction = TITS;
	else if(firstlast == "ck") last_instruction = COCK;
	else if(firstlast == "kb") last_instruction = KNOB;
	else if(firstlast == "ae") last_instruction = ARSE;
	else if(firstlast == "bt") last_instruction = BUTT;
	else throw std::runtime_error("unknown f*cking instruction: '"
				      + firstlast.insert(1, "**") + "'");
	return last_instruction;
      }
    }
  }
}
  
void FFInterpreter::execute_instruction(Program::const_iterator &ip) {
  //  print_instruction(err, *ip, false);
  switch(*(ip)) {
  case FUCK: tm.up(); break;
  case SHAG: tm.down(); break;
  case BOOB: tm.inc(); break;
  case TITS: tm.dec(); break;
  case COCK: out << tm.peek() << std::flush; break;
  case KNOB: tm.poke((unsigned char) get(in)); break;
  case ARSE:
    if(!tm.peek()) {
      ip++;
      int level = 0;
      while(ip != program.end() && (level || *ip != BUTT)) {
	if(*ip == ARSE) level++;
	else if(*ip == BUTT) level--;
	ip++;
      }
      if(ip == program.end())
	throw std::runtime_error("unmatched a**e");
    }
    break;
  case BUTT: 
    {
      ip--;
      int level = 0;
      while(ip != program.begin() && (level || *ip != ARSE)) {
	if(*ip == BUTT) level++;
	else if(*ip == ARSE) level--;
	ip--;
      }
      if(level || *ip != ARSE)
	throw std::runtime_error("unmatched b**t");
      return; // dont increase the ip
      break;
    }
  default: break;
  }
  ip++;
}

void FFInterpreter::print_instruction(std::ostream &out, instruction inst, 
				      bool censor) 
{
  std::string name;
  switch(inst) {
  case FUCK: name = "fuck"; break;
  case SHAG: name = "shag"; break;
  case BOOB: name = "boob"; break;
  case TITS: name = "tits"; break;
  case COCK: name = "cock"; break;
  case KNOB: name = "knob"; break;
  case ARSE: name = "arse"; break;
  case BUTT: name = "butt"; break;
  default: name = "????"; break;
  }
  if(censor) {
    name[1]='*';
    name[2]='*';
  }
  out << name << " ";
}
