#ifndef TOURETTEMACHINE
#define TOURETTEMACHINE
#include <map>

// Turing machine for f*ckf*ck by Fredrik Arnerup

class TouretteMachine {
private:
  // Since most f*ckf*ck programs probably don't need much memory,
  // using a map will save a lot
  typedef std::map<long, unsigned char> Memory;
  Memory memory;
  long pointer;
  
  void set(long address, unsigned char value) {
    memory[address] = value;
  }

  unsigned char get(long address) const {
    Memory::const_iterator i = memory.find(address);
    return i != memory.end() ? i->second : 0;
  }

public:

  // move pointer forward
  void up() {pointer++;}

  // move pointer backwards
  void down() {pointer--;}

  // increment value at pointer
  void inc() {set(pointer, get(pointer) + 1);}

  // decrement value at pointer
  void dec() {set(pointer, get(pointer) - 1);}

  // read value att pointer
  unsigned char peek() const {return get(pointer);}

  // set value att pointer
  void poke(unsigned char c) {set(pointer, c);}
};

#endif
